#include <SPI.h>
#include <SoftwareSerial.h>
#include <TimeLib.h>
#include <TinyGPS.h>

#define cwSpace 4     // The CW space or "no carrier" tone
#define cwMark 5      // The CW mark or carrier tone
#define ledGPS 2      // TX Led
#define RX 5          // GPS Module TXD
#define TX -1          // GPS Module RXD
#define PI4_DELAY 167 // PI4 Delay


// Change this block to match the messages
//                    12345678
char pi4Message[9] = "RA4FHE  ";                             // The PI4 message. Must be 8 charaters, left justified and padded with spaces if needed
char call[] = "RA4FHE";                                      // The CW call. Only A-Z, 0-9 and /
char locator[] = "LO23MF";                                   // The CW locator. Only A-X, 0-9 and /
uint8_t convert[] = { 0x62, 0x66, 0x6A, 0x6E, 0x72, 0x76 };  // FSK_DEV_SEL R34
static byte symbolNo;
byte symbols[146];
byte callLength;
byte locatorLength;

EspSoftwareSerial::UART gnssSerial;
TinyGPS gps;

byte data[]{
  0x3C, 0xA0, 0x00,
  0x3A, 0x8C, 0x00,
  0x35, 0x78, 0x06,
  0x2F, 0x40, 0x00,
  0x2E, 0x00, 0x1D,
  0x2A, 0x00, 0x10,
  0x29, 0x08, 0x18,
  0x28, 0x18, 0x5C,
  0x27, 0x11, 0xFB,
  0x23, 0x00, 0x03,
  0x22, 0x00, 0x62,
  0x21, 0x00, 0x00,
  0x20, 0x00, 0x00,
  0x1F, 0x00, 0x00,
  0x1E, 0x00, 0x00,
  0x1D, 0x00, 0x00,
  0x1C, 0x00, 0x13,
  0x1B, 0x00, 0x0E,
  0x1A, 0x00, 0x09,
  0x19, 0x00, 0x05,
  0x18, 0x04, 0x04,
  0x17, 0x04, 0x04,
  0x16, 0x85, 0x81,
  0x15, 0x01, 0x01,
  0x14, 0x30, 0x53,
  0x13, 0xFF, 0xFF,
  0x12, 0x87, 0xCD,
  0x11, 0xFF, 0x22,
  0x10, 0x00, 0x00,
  0x0F, 0x00, 0x00,
  0x0E, 0x00, 0x00,
  0x0D, 0xF5, 0xEA,
  0x0C, 0x17, 0xA2,
  0x0B, 0x0E, 0x2E,
  0x0A, 0x04, 0xBA,
  0x09, 0xFB, 0x44,
  0x08, 0x04, 0x04,
  0x07, 0x04, 0x84,
  0x06, 0x8C, 0x81,
  0x05, 0x01, 0x01,
  0x04, 0x30, 0x58,
  0x03, 0xFF, 0xFF,
  0x02, 0xBD, 0x33,  // 144.4709999768
  0x01, 0xFF, 0xE7,
  0x00, 0x02, 0x01
};

void syncGps() {

  while (gnssSerial.available()) {
    if (gps.encode(gnssSerial.read())) {
      unsigned long age;
      int year;
      byte month, day, hour, minute, second;
      gps.crack_datetime(&year, &month, &day, &hour, &minute, &second, NULL, &age);
      if (age < 500) {
        setTime(hour, minute, second, day, month, year);
      }
    }
  }
}


byte CharToMorse(char ch) {  // e.g. 0x42 = 01000010 = 01000 010 = .- 2 bits long = A
  const byte letters[] = { 0x42, 0x84, 0xA4, 0x83, 0x01, 0x24, 0xC3, 0x04, 0x02, 0x74, 0xA3, 
                           0x44, 0xC2, 0x82, 0xE3, 0x64, 0xD4, 0x43, 0x03, 0x81, 0x23, 0x14,
                           0x63, 0x94, 0xB4, 0xC4 };
  const byte numbers[] = { 0x95, 0xFD, 0x7D, 0x3D, 0x1D, 0x0D, 0x05, 0x85, 0xC5, 0xE5, 0xF5 };

  if (ch == 0)
    return (0);       // Not a valid morse character
  else if (ch < 'A')  // Get then Morse pattern
    return (numbers[ch - '/']);
  else
    return (letters[ch - 'A']);
}

void SendMorse(const char *info, const byte msgLen) {
  byte i, j;
  byte morse;
  byte morseLength;

  for (i = 0; i < msgLen; i++) {
    morse = info[i];
    morseLength = morse & 0x07;  // Bit2 to Bit0 of morse is the length

    for (j = 0; j < morseLength; j++) {
      sendData(convert[cwMark]);
      Serial.println(convert[cwMark], HEX);
      if ((morse & 0x80) == 0x80)  // If MSB 0 = dot, 1 = dash,
        delay(300);                // It is a dash, so wait 3 dot durations
      else
        delay(100);  // It is a dot, so wait 1 dot duration

      sendData(convert[cwSpace]);
       Serial.println(convert[cwSpace], HEX);
      delay(100);
      morse = morse << 1;  // Point to next bit
    }
    delay(200);  // Inter morse character pause, 1 dot duration already "gone" so only 2 left
  }
}

//     Do NOT modify unless you really know what you are doing!
//
//                               _123456789_123456789_123456789_1234567
const char pi4Chars[] PROGMEM = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ /";
const byte pi4Vector[] PROGMEM = {
  0, 0, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1,
  1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0,
  0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0,
  1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1,
  0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 1
};

byte Paritypi(uint32_t value) {
  byte even = 0;
  byte bitNo;

  for (bitNo = 0; bitNo < 32; bitNo++)
    if (((value >> bitNo) & 0x01) != 0)
      even = 1 - even;
  return even;
}


void PI4MakeMessage(char *msg) {
  // Source encoding
  uint64_t sourceEnc = 0;

  for (byte i = 0; i < 8; i++)
    sourceEnc = sourceEnc * 38 + (uint64_t)(strchr(pi4Chars, msg[i]) - pi4Chars);
  // Convolutional Encoding
  const uint32_t poly1 = 0xF2D05351;
  const uint32_t poly2 = 0xE4613C47;
  uint32_t n = 0;
  byte convEnc[146] = { 0 };

  byte i = 0;
  for (byte j = 0; j < 73; j++) {
    n <<= 1;
    if ((sourceEnc & 0x20000000000LLU) != 0)
      n |= 1;
    sourceEnc <<= 1;

    convEnc[i++] = Paritypi(n & poly1);
    convEnc[i++] = Paritypi(n & poly2);
  }

  // Interleaving
  byte p = 0;
  byte r = 0;
  byte interleaved[146] = { 0 };

  for (int i = 0; i <= 255; i++) {
    for (byte j = 0; j < 8; j++) {
      if (((i >> j) & 0x01) == 0x01)
        r |= 1 << (7 - j);
      else
        r &= ~(1 << (7 - j));
    }

    if ((p < 146) && (r < 146))
      interleaved[r] = convEnc[p++];
  }

  //  Merge with Sync Vector
  for (byte i = 0; i < 146; i++)
    symbols[i] = pgm_read_byte(&pi4Vector[i]) + 2 * interleaved[i];
}



void sendData(uint8_t data) {
  digitalWrite(SS, LOW);
  SPI.transfer(0x22);
  SPI.transfer(0x00);
  SPI.transfer(data);  
  SPI.endTransaction();
  digitalWrite(SS, HIGH);
}

void SendPI4() {
  uint32_t period = millis() - PI4_DELAY;
  symbolNo = 0;
  while (symbolNo < 146) {
    uint32_t curMill = millis();
    if (curMill - period >= PI4_DELAY) {
      period = curMill;
      sendData(convert[symbols[symbolNo]]);
      Serial.println(convert[symbols[symbolNo]], HEX);
      symbolNo++;
    }
  }
}

void setup() {

  pinMode(ledGPS, OUTPUT);
  digitalWrite(ledGPS, HIGH);
  Serial.begin(115200);
  gnssSerial.begin(9600, EspSoftwareSerial::SWSERIAL_8N1, RX, TX);
  SPI.begin();
  pinMode(SS, HIGH);
  delay(50);
  SPI.beginTransaction(SPISettings(1000000, MSBFIRST, SPI_MODE3));
  for (int i = 0; i < sizeof(data); i += 3) {
    digitalWrite(SS, LOW);
    SPI.transfer(char(data[i]));
    SPI.transfer(char(data[i + 1]));
    SPI.transfer(char(data[i + 2]));
    SPI.endTransaction();
    digitalWrite(SS, HIGH);
  }

  byte i;

  PI4MakeMessage(pi4Message);       // Convert the PI4 message to symbols
  callLength = strlen(call);        // Find the length of the CW call
  for (i = 0; i < callLength; i++)  // Convert call to Morse
    call[i] = CharToMorse(call[i]);
  locatorLength = strlen(locator);     // Find the length of the CW locator
  for (i = 0; i < locatorLength; i++)  // Convert locator to Morse
    locator[i] = CharToMorse(locator[i]);
}

void loop() {
  syncGps();
  sendData(convert[cwMark]);
  if (second() == 0) {
    SendPI4();
    sendData(convert[cwSpace]);
    delay(667);
    SendMorse(call, callLength);
    delay(500); 
    SendMorse(locator, locatorLength);
    delay(500); 
  }              
}